"""Unit tests for timestreaminfluxdb-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


"""Unit tests for timestreaminfluxdb-supported APIs."""


@mock_aws
def test_create_db_instance_success():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    assert response["allocatedStorage"] == 123
    assert response["dbInstanceType"] == "db.influx.medium"
    assert response["name"] == "test-instance"
    assert response["vpcSecurityGroupIds"] == ["sg-0123456789abcdef0"]
    assert response["vpcSubnetIds"] == ["subnet-0123456789abcdef0"]

    assert "arn" in response
    assert "availabilityZone" in response
    assert "dbStorageType" in response
    assert "deploymentType" in response
    assert "endpoint" in response
    assert "id" in response
    assert "influxAuthParametersSecretArn" in response
    assert "networkType" in response
    assert "port" in response
    assert "publiclyAccessible" in response
    assert "secondaryAvailabilityZone" in response
    assert "status" in response
    assert "logDeliveryConfiguration" in response


@mock_aws
def test_create_db_instance_duplicate_identifier():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    with pytest.raises(ClientError) as exc:
        client.create_db_instance(
            name="test-instance",
            password="password123",
            dbInstanceType="db.influx.medium",
            vpcSubnetIds=["subnet-0123456789abcdef0"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            allocatedStorage=123,
        )

    assert exc.value.response["Error"]["Code"] == "ConflictException"


@mock_aws
def test_create_db_instance_name_invalid():
    invalid_names = [
        "1muststartwithletter",
        "cantendwithhyphen_",
        "nodoublehyphen--",
        "no",
        "longlonglonglonglonglonglonglonglonglongname",
    ]
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    for invalid_name in invalid_names:
        with pytest.raises(ClientError) as exc:
            client.create_db_instance(
                name="invalid_name",
                password="password123",
                dbInstanceType="db.influx.medium",
                vpcSubnetIds=["subnet-0123456789abcdef0"],
                vpcSecurityGroupIds=["sg-0123456789abcdef0"],
                allocatedStorage=123,
            )
        assert exc.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_create_db_instance_invalid_storage_type():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.create_db_instance(
            name="test-instance",
            password="password123",
            dbInstanceType="db.influx.medium",
            vpcSubnetIds=["subnet-0123456789abcdef0"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            allocatedStorage=123,
            dbStorageType="invalid",
        )

        assert exc.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_create_db_instance_invalid_instance_type():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        client.create_db_instance(
            name="test-instance",
            password="password123",
            vpcSubnetIds=["subnet-0123456789abcdef0"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            allocatedStorage=123,
            dbInstanceType="invalid",
        )

    assert exc.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_delete_db_instance():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    id = response["id"]

    response = client.delete_db_instance(identifier=id)
    assert response["allocatedStorage"] == 123
    assert response["dbInstanceType"] == "db.influx.medium"
    assert response["name"] == "test-instance"
    assert response["vpcSecurityGroupIds"] == ["sg-0123456789abcdef0"]
    assert response["vpcSubnetIds"] == ["subnet-0123456789abcdef0"]

    assert "arn" in response
    assert "availabilityZone" in response
    assert "dbStorageType" in response
    assert "deploymentType" in response
    assert "endpoint" in response
    assert "id" in response
    assert "influxAuthParametersSecretArn" in response
    assert "networkType" in response
    assert "port" in response
    assert "publiclyAccessible" in response
    assert "secondaryAvailabilityZone" in response
    assert "logDeliveryConfiguration" in response

    assert response["status"] == "DELETING"


@mock_aws
def test_delete_db_instance_invalid_name():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    id = "-100000"
    with pytest.raises(ClientError) as exc:
        client.delete_db_instance(identifier=id)

    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_get_db_instance():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    created_instance = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    id = created_instance["id"]
    response = client.get_db_instance(identifier=id)
    assert response["allocatedStorage"] == 123
    assert response["dbInstanceType"] == "db.influx.medium"
    assert response["name"] == "test-instance"
    assert response["vpcSecurityGroupIds"] == ["sg-0123456789abcdef0"]
    assert response["vpcSubnetIds"] == ["subnet-0123456789abcdef0"]

    assert "arn" in response
    assert "availabilityZone" in response
    assert "dbStorageType" in response
    assert "deploymentType" in response
    assert "endpoint" in response
    assert "id" in response
    assert "influxAuthParametersSecretArn" in response
    assert "networkType" in response
    assert "port" in response
    assert "publiclyAccessible" in response
    assert "secondaryAvailabilityZone" in response
    assert "logDeliveryConfiguration" in response
    assert "status" in response


@mock_aws
def test_get_db_instance_invalid():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    id = "-100000"
    with pytest.raises(ClientError) as exc:
        client.get_db_instance(identifier=id)

    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_db_instances_empty():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")
    response = client.list_db_instances()
    assert len(response["items"]) == 0


@mock_aws
def test_list_db_instances():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")
    client.create_db_instance(
        name="test-instance1",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    client.create_db_instance(
        name="test-instance2",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    response = client.list_db_instances()
    resources = response["items"]
    assert len(resources) == 2

    for resource in resources:
        assert "allocatedStorage" in resource
        assert "arn" in resource
        assert "dbInstanceType" in resource
        assert "dbStorageType" in resource
        assert "deploymentType" in resource
        assert "endpoint" in resource
        assert "id" in resource
        assert "name" in resource
        assert "networkType" in resource
        assert "port" in resource
        assert "status" in resource

    # delete one and confirm only one left
    client.delete_db_instance(identifier=resources[0]["id"])

    response = client.list_db_instances()
    resources = response["items"]
    assert len(resources) == 1



from __future__ import annotations

from typing import Set

from antlr4 import RecognitionException

from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service import (
    StateTaskService,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_api_gateway import (
    StateTaskServiceApiGateway,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_aws_sdk import (
    StateTaskServiceAwsSdk,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_batch import (
    StateTaskServiceBatch,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_dynamodb import (
    StateTaskServiceDynamoDB,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_ecs import (
    StateTaskServiceEcs,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_events import (
    StateTaskServiceEvents,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_glue import (
    StateTaskServiceGlue,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_lambda import (
    StateTaskServiceLambda,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_sfn import (
    StateTaskServiceSfn,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_sns import (
    StateTaskServiceSns,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_sqs import (
    StateTaskServiceSqs,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_unsupported import (
    StateTaskServiceUnsupported,
)

_UNSUPPORTED_SERVICE_NAMES: Set[str] = {
    "athena",
    "bedrock",
    "codebuild",
    "eks",
    "elasticmapreduce",
    "emr-containers",
    "emr-serverless",
    "databrew",
    "mediaconvert",
    "sagemaker",
}


# TODO: improve on factory constructor (don't use SubtypeManager: cannot reuse state task instances).
def state_task_service_for(service_name: str) -> StateTaskService:
    if service_name == "aws-sdk":
        return StateTaskServiceAwsSdk()
    if service_name == "lambda":
        return StateTaskServiceLambda()
    if service_name == "sqs":
        return StateTaskServiceSqs()
    if service_name == "states":
        return StateTaskServiceSfn()
    if service_name == "dynamodb":
        return StateTaskServiceDynamoDB()
    if service_name == "apigateway":
        return StateTaskServiceApiGateway()
    if service_name == "sns":
        return StateTaskServiceSns()
    if service_name == "events":
        return StateTaskServiceEvents()
    if service_name == "ecs":
        return StateTaskServiceEcs()
    if service_name == "glue":
        return StateTaskServiceGlue()
    if service_name == "batch":
        return StateTaskServiceBatch()
    if service_name in _UNSUPPORTED_SERVICE_NAMES:
        return StateTaskServiceUnsupported()
    raise RecognitionException(f"Unknown service '{service_name}'")



import boto3

from moto import mock_aws

from .test_helper_functions import CREATE_WEB_ACL_BODY


@mock_aws
def test_list_tags_for_resource__none_supplied():
    conn = boto3.client("wafv2", region_name="us-east-1")
    arn = conn.create_web_acl(**CREATE_WEB_ACL_BODY("John", "REGIONAL"))["Summary"][
        "ARN"
    ]

    tag_info = conn.list_tags_for_resource(ResourceARN=arn)["TagInfoForResource"]
    assert tag_info["TagList"] == []


@mock_aws
def test_list_tags_for_resource():
    conn = boto3.client("wafv2", region_name="us-east-1")
    arn = conn.create_web_acl(
        Name="test",
        Scope="CLOUDFRONT",
        DefaultAction={"Allow": {}},
        Tags=[{"Key": f"k{idx}", "Value": "v1"} for idx in range(10)],
        VisibilityConfig={
            "SampledRequestsEnabled": False,
            "CloudWatchMetricsEnabled": False,
            "MetricName": "idk",
        },
    )["Summary"]["ARN"]

    tag_info = conn.list_tags_for_resource(ResourceARN=arn)["TagInfoForResource"]
    assert len(tag_info["TagList"]) == 10

    page1 = conn.list_tags_for_resource(ResourceARN=arn, Limit=2)
    assert len(page1["TagInfoForResource"]["TagList"]) == 2

    page2 = conn.list_tags_for_resource(
        ResourceARN=arn, Limit=3, NextMarker=page1["NextMarker"]
    )
    assert len(page2["TagInfoForResource"]["TagList"]) == 3

    page3 = conn.list_tags_for_resource(ResourceARN=arn, NextMarker=page2["NextMarker"])
    assert len(page3["TagInfoForResource"]["TagList"]) == 5


@mock_aws
def test_tag_resource():
    conn = boto3.client("wafv2", region_name="us-east-1")
    arn = conn.create_web_acl(
        Name="test",
        Scope="CLOUDFRONT",
        DefaultAction={"Allow": {}},
        Tags=[{"Key": "k1", "Value": "v1"}],
        VisibilityConfig={
            "SampledRequestsEnabled": False,
            "CloudWatchMetricsEnabled": False,
            "MetricName": "idk",
        },
    )["Summary"]["ARN"]

    conn.tag_resource(
        ResourceARN=arn,
        Tags=[
            {"Key": "k2", "Value": "v2"},
        ],
    )

    tag_info = conn.list_tags_for_resource(ResourceARN=arn)["TagInfoForResource"]
    assert tag_info["TagList"] == [
        {"Key": "k1", "Value": "v1"},
        {"Key": "k2", "Value": "v2"},
    ]


@mock_aws
def test_untag_resource():
    conn = boto3.client("wafv2", region_name="us-east-1")
    arn = conn.create_web_acl(
        Name="test",
        Scope="CLOUDFRONT",
        DefaultAction={"Allow": {}},
        Tags=[{"Key": "k1", "Value": "v1"}, {"Key": "k2", "Value": "v2"}],
        VisibilityConfig={
            "SampledRequestsEnabled": False,
            "CloudWatchMetricsEnabled": False,
            "MetricName": "idk",
        },
    )["Summary"]["ARN"]

    conn.untag_resource(ResourceARN=arn, TagKeys=["k1"])

    tag_info = conn.list_tags_for_resource(ResourceARN=arn)["TagInfoForResource"]
    assert tag_info["TagList"] == [{"Key": "k2", "Value": "v2"}]



import csv
import io
import itertools
from collections import OrderedDict

from moto.stepfunctions.parser.api import HistoryEventType, MapRunFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_output_transformer.transformer import (
    ResourceOutputTransformer,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.reader_config_decl import (
    CSVHeaderLocationOutput,
    ReaderConfigOutput,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails


class ResourceOutputTransformerCSV(ResourceOutputTransformer):
    def _eval_body(self, env: Environment) -> None:
        reader_config: ReaderConfigOutput = env.stack.pop()
        resource_value: str = env.stack.pop()

        csv_file = io.StringIO(resource_value)
        csv_reader = csv.reader(csv_file)

        max_items: int = reader_config["MaxItemsValue"]
        csv_reader_slice = itertools.islice(csv_reader, max_items)

        location = reader_config["CSVHeaderLocation"]
        if location == CSVHeaderLocationOutput.FIRST_ROW:
            headers = next(csv_reader)
        elif location == CSVHeaderLocationOutput.GIVEN:
            headers = reader_config["CSVHeaders"]
        else:
            raise ValueError(f"Unknown CSVHeaderLocation value '{location}'.")

        if len(set(headers)) < len(headers):
            error_name = StatesErrorName(typ=StatesErrorNameType.StatesItemReaderFailed)
            failure_event = FailureEvent(
                env=env,
                error_name=error_name,
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    mapRunFailedEventDetails=MapRunFailedEventDetails(
                        error=error_name.error_name,
                        cause="CSV headers cannot contain duplicates.",
                    )
                ),
            )
            raise FailureEventException(failure_event=failure_event)

        transformed_outputs = list()
        for row in csv_reader_slice:
            transformed_output = dict()
            for i, header in enumerate(headers):
                transformed_output[header] = row[i] if i < len(row) else ""
            transformed_outputs.append(
                OrderedDict(
                    sorted(
                        transformed_output.items(),
                        key=lambda item: (item[0].isalpha(), item[0]),
                    )
                )
            )

        env.stack.append(transformed_outputs)



"""Unit tests for kms-supported APIs."""

import boto3
import pytest

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_rotate_key_on_demand_with_existing_key():
    client = boto3.client("kms", region_name="us-east-2")

    key_id = client.create_key(Policy="my policy")["KeyMetadata"]["KeyId"]

    resp = client.rotate_key_on_demand(KeyId=key_id)

    assert resp["KeyId"] == key_id


@mock_aws
def test_rotate_key_on_demand_with_non_existing_key_fails():
    client = boto3.client("kms", region_name="us-east-2")

    with pytest.raises(client.exceptions.NotFoundException):
        client.rotate_key_on_demand(KeyId="some-id")


@mock_aws
def test_list_key_rotations_with_non_existing_key_fails():
    client = boto3.client("kms", region_name="us-east-2")

    with pytest.raises(client.exceptions.NotFoundException):
        client.list_key_rotations(KeyId="some-id")


@mock_aws
def test_list_key_rotations_are_empty_on_new_key():
    client = boto3.client("kms", region_name="us-east-2")

    key_id = client.create_key(Policy="my policy")["KeyMetadata"]["KeyId"]

    resp = client.list_key_rotations(KeyId=key_id)

    assert len(resp["Rotations"]) == 0
    assert resp["Truncated"] is False
    assert "NextMarker" not in resp


@mock_aws
def test_list_key_rotations_returns_one_rotation():
    # Arrange
    client = boto3.client("kms", region_name="us-east-2")
    key_id = client.create_key(Policy="my policy")["KeyMetadata"]["KeyId"]
    client.rotate_key_on_demand(KeyId=key_id)

    # Act
    resp = client.list_key_rotations(KeyId=key_id)

    # Assert
    assert len(resp["Rotations"]) == 1
    assert resp["Truncated"] is False
    assert "NextMarker" not in resp
    assert resp["Rotations"][0]["RotationType"] == "ON_DEMAND"


@mock_aws
def test_list_key_rotations_returns_truncated_and_next_marker():
    # Arrange
    client = boto3.client("kms", region_name="us-east-2")
    key_id = client.create_key(Policy="my policy")["KeyMetadata"]["KeyId"]
    client.rotate_key_on_demand(KeyId=key_id)
    client.rotate_key_on_demand(KeyId=key_id)
    client.rotate_key_on_demand(KeyId=key_id)

    # Act
    resp = client.list_key_rotations(KeyId=key_id, Limit=1)

    # Assert
    assert len(resp["Rotations"]) == 1
    assert resp["Truncated"] is True
    assert "NextMarker" in resp


@mock_aws
def test_list_key_rotations_pagination():
    # Arrange
    client = boto3.client("kms", region_name="us-east-2")
    key_id = client.create_key(Policy="my policy")["KeyMetadata"]["KeyId"]
    client.rotate_key_on_demand(KeyId=key_id)
    client.rotate_key_on_demand(KeyId=key_id)
    client.rotate_key_on_demand(KeyId=key_id)
    initial_page = client.list_key_rotations(KeyId=key_id, Limit=1)

    # Act

    final_page = client.list_key_rotations(
        KeyId=key_id, Limit=2, Marker=initial_page["NextMarker"]
    )

    # Assert
    assert len(final_page["Rotations"]) == 2
    assert final_page["Truncated"] is False
    assert "NextMarker" not in final_page



"""TimestreamInfluxDBBackend class with methods for supported APIs."""

from enum import Enum
from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.tagging_service import TaggingService

from .exceptions import (
    ConflictException,
    ResourceNotFoundException,
    ValidationException,
)
from .utils import random_id, validate_name


class InstanceStatus(str, Enum):
    CREATING = "CREATING"
    AVAILABLE = "AVAILABLE"
    DELETING = "DELETING"
    MODIFYING = "MODIFYING"
    UPDATING = "UPDATING"
    DELETED = "DELETED"
    FAILED = "FAILED"
    UPDATING_DEPLOYMENT_TYPE = "UPDATING_DEPLOYMENT_TYPE"
    UPDATING_INSTANCE_TYPE = "UPDATING_INSTANCE_TYPE"


class NetworkType(str, Enum):
    IPV4 = "IPV4"
    DUAL = "DUAL"


class InstanceType(str, Enum):
    DB_INFLUX_MEDIUM = "db.influx.medium"
    DB_INFLUX_LARGE = "db.influx.large"
    DB_INFLUX_XLARGE = "db.influx.xlarge"
    DB_INFLUX_2XLARGE = "db.influx.2xlarge"
    DB_INFLUX_4XLARGE = "db.influx.4xlarge"
    DB_INFLUX_8XLARGE = "db.influx.8xlarge"
    DB_INFLUX_12XLARGE = "db.influx.12xlarge"
    DB_INFLUX_16XLARGE = "db.influx.16xlarge"


class DBStorageType(str, Enum):
    InfluxIOIncludedT1 = "InfluxIOIncludedT1"
    InfluxIOIncludedT2 = "InfluxIOIncludedT2"
    InfluxIOIncludedT3 = "InfluxIOIncludedT3"


class DeploymentType(str, Enum):
    SINGLE_AZ = "SINGLE_AZ"
    WITH_MULTIAZ_STANDBY = "WITH_MULTIAZ_STANDBY"


class DBInstance(BaseModel):
    def __init__(
        self,
        name: str,
        username: Optional[str],
        password: str,
        organization: str,
        bucket: str,
        dbInstanceType: str,
        vpcSubnetIds: List[str],
        vpcSecurityGroupIds: List[str],
        publiclyAccessible: bool,
        dbStorageType: str,
        allocatedStorage: int,
        dbParameterGroupIdentifier: Optional[str],
        deploymentType: str,
        logDeliveryConfiguration: Optional[Dict[str, Any]],
        tags: Optional[Dict[str, Any]],
        port: int,
        networkType: str,
        region_name: str,
        account_id: str,
        endpoint_id: str,
    ):
        # Generate a random id of size 10
        self.id = random_id()

        self.name = name
        self.username = username
        self.password = password
        self.organization = organization
        self.bucket = bucket
        self.db_instance_type = dbInstanceType
        self.vpc_subnet_ids = vpcSubnetIds
        self.vpc_security_group_ids = vpcSecurityGroupIds
        self.publicly_accessible = publiclyAccessible
        self.db_storage_type = dbStorageType
        self.allocated_storage = allocatedStorage
        self.db_parameter_group_id = dbParameterGroupIdentifier
        self.deployment_type = deploymentType
        self.log_delivery_configuration = logDeliveryConfiguration
        self.port = port
        self.network_type = networkType
        self.status = InstanceStatus.CREATING
        self.arn = f"arn:aws:timestream-influxdb:{region_name}:{account_id}:db-instance/{self.id}"
        self.endpoint = (
            f"{self.id}-{endpoint_id}.timestream-influxdb.{region_name}.on.aws"
        )
        # Before 12/09/2024, there was a different endpoint format.
        self.endpoint_old = (
            f"{self.name}-{endpoint_id}.timestream-influxdb.{region_name}.on.aws"
        )

        self.availability_zone = ""  # TODO implement this
        self.secondary_availability_zone = ""  # TODO implement this

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "arn": self.arn,
            "status": self.status,
            "endpoint": self.endpoint,
            "port": self.port,
            "networkType": self.network_type,
            "dbInstanceType": self.db_instance_type,
            "dbStorageType": self.db_storage_type,
            "allocatedStorage": self.allocated_storage,
            "deploymentType": self.deployment_type,
            "vpcSubnetIds": self.vpc_subnet_ids,
            "publiclyAccessible": self.publicly_accessible,
            "vpcSecurityGroupIds": self.vpc_security_group_ids,
            "dbParameterGroupIdentifier": self.db_parameter_group_id,  # TODO implement this
            "availabilityZone": self.availability_zone,  # TODO implement this
            "secondaryAvailabilityZone": self.secondary_availability_zone,  # TODO implement this
            "logDeliveryConfiguration": self.log_delivery_configuration,  # TODO implement this
            "influxAuthParametersSecretArn": "",  # TODO implement this
        }


class TimestreamInfluxDBBackend(BaseBackend):
    """Implementation of TimestreamInfluxDB APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)

        # the endpoint identifier is unique per account and per region
        # https://docs.aws.amazon.com/timestream/latest/developerguide/timestream-for-influxdb.html
        self.endpoint_id: str = random_id(10)
        self.db_instances: Dict[str, DBInstance] = {}
        self.tagger = TaggingService()

    def create_db_instance(
        self,
        name: str,
        username: Optional[str],  # required if using InfluxDB UI though
        password: str,
        organization: str,
        bucket: str,
        db_instance_type: str,
        vpc_subnet_ids: List[str],
        vpc_security_group_ids: List[str],
        db_storage_type: str,
        publicly_accessible: bool,
        allocated_storage: int,
        db_parameter_group_identifier: str,
        deployment_type: str,
        log_delivery_configuration: Optional[Dict[str, Any]],
        tags: Optional[Dict[str, str]],
        port: int,
        network_type: str,
    ) -> DBInstance:
        """
        dbParameterGroupIdentifier argument is not yet handled
        deploymentType currently is auto set to 'SINGLE_AZ' if not passed in.
        publicAccessible is not yet handled
        logDeliveryConfiguration is not yet handled
        AvailabilityZone and SecondaryAvailabilityZone are not yet handled
        influxAuthParametersSecretArn is not yet handled
        """

        # Checks:
        for db_instance in self.db_instances.values():
            if db_instance.name == name:
                raise ConflictException(
                    f"A DB Instance with the name {name} already exists"
                )

        validate_name(name)

        if db_storage_type not in [t.value for t in DBStorageType]:
            raise ValidationException(f"Unknown DB storage type {db_storage_type}")

        if db_instance_type not in [t.value for t in InstanceType]:
            raise ValidationException(f"Unknown DB instance type {db_instance_type}")

        new_instance = DBInstance(
            name,
            username,
            password,
            organization,
            bucket,
            db_instance_type,
            vpc_subnet_ids,
            vpc_security_group_ids,
            publicly_accessible,
            db_storage_type,
            allocated_storage,
            db_parameter_group_identifier,
            deployment_type,
            log_delivery_configuration,
            tags,
            port,
            network_type,
            self.region_name,
            self.account_id,
            self.endpoint_id,
        )

        # add to the list
        self.db_instances[new_instance.id] = new_instance

        # add tags
        if tags:
            self.tag_resource(new_instance.arn, tags)

        return new_instance

    def delete_db_instance(self, id: str) -> DBInstance:
        if id not in self.db_instances:
            raise ResourceNotFoundException(f"DB Instance with id {id} not found")

        # mark as deleting
        self.db_instances[id].status = InstanceStatus.DELETING
        return self.db_instances.pop(id)

    def get_db_instance(self, id: str) -> DBInstance:
        if id not in self.db_instances:
            raise ResourceNotFoundException(f"DB Instance with id {id} not found")

        return self.db_instances[id]

    def list_db_instances(self) -> List[Dict[str, Any]]:
        """
        Pagination is not yet implemented
        """
        return [
            {
                "allocatedStorage": instance.allocated_storage,
                "arn": instance.arn,
                "dbInstanceType": instance.db_instance_type,
                "dbStorageType": instance.db_storage_type,
                "deploymentType": instance.deployment_type,
                "endpoint": instance.endpoint,
                "id": instance.id,
                "name": instance.name,
                "networkType": instance.network_type,
                "port": instance.port,
                "status": instance.status,
            }
            for instance in self.db_instances.values()
        ]

    def tag_resource(self, resource_arn: str, tags: Dict[str, str]) -> None:
        tag_list = self.tagger.convert_dict_to_tags_input(tags)
        errmsg = self.tagger.validate_tags(tag_list)
        if errmsg:
            raise ValidationException(errmsg)
        self.tagger.tag_resource(resource_arn, tag_list)

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def list_tags_for_resource(self, resource_arn: str) -> Dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)


timestreaminfluxdb_backends = BackendDict(
    TimestreamInfluxDBBackend,
    "timestream-influxdb",
    additional_regions=[
        "us-east-1",
        "us-east-2",
        "us-west-2",
        "eu-central-1",
        "eu-west-1",
        "ap-southeast-2",
        "ap-northeast-1",
    ],
)



from __future__ import annotations

from enum import Enum
from typing import Dict, Final

from moto.stepfunctions.parser.asl.antlr.runtime.ASLLexer import ASLLexer


class StatesErrorNameType(Enum):
    StatesALL = ASLLexer.ERRORNAMEStatesALL
    StatesStatesDataLimitExceeded = ASLLexer.ERRORNAMEStatesDataLimitExceeded
    StatesHeartbeatTimeout = ASLLexer.ERRORNAMEStatesHeartbeatTimeout
    StatesTimeout = ASLLexer.ERRORNAMEStatesTimeout
    StatesTaskFailed = ASLLexer.ERRORNAMEStatesTaskFailed
    StatesPermissions = ASLLexer.ERRORNAMEStatesPermissions
    StatesResultPathMatchFailure = ASLLexer.ERRORNAMEStatesResultPathMatchFailure
    StatesParameterPathFailure = ASLLexer.ERRORNAMEStatesParameterPathFailure
    StatesBranchFailed = ASLLexer.ERRORNAMEStatesBranchFailed
    StatesNoChoiceMatched = ASLLexer.ERRORNAMEStatesNoChoiceMatched
    StatesIntrinsicFailure = ASLLexer.ERRORNAMEStatesIntrinsicFailure
    StatesExceedToleratedFailureThreshold = (
        ASLLexer.ERRORNAMEStatesExceedToleratedFailureThreshold
    )
    StatesItemReaderFailed = ASLLexer.ERRORNAMEStatesItemReaderFailed
    StatesResultWriterFailed = ASLLexer.ERRORNAMEStatesResultWriterFailed
    StatesRuntime = ASLLexer.ERRORNAMEStatesRuntime
    StatesQueryEvaluationError = ASLLexer.ERRORNAMEStatesQueryEvaluationError

    def to_name(self) -> str:
        return _error_name(self)

    @classmethod
    def from_name(cls, name: str) -> StatesErrorNameType:
        error_name = _REVERSE_NAME_LOOKUP.get(name, None)
        if error_name is None:
            raise ValueError(f"Unknown ErrorName type, got: '{name}'.")
        return cls(error_name.value)


def _error_name(error_name: StatesErrorNameType) -> str:
    return ASLLexer.literalNames[error_name.value][2:-2]


def _reverse_error_name_lookup() -> Dict[str, StatesErrorNameType]:
    lookup: Dict[str, StatesErrorNameType] = dict()
    for error_name in StatesErrorNameType:
        error_text: str = _error_name(error_name)
        lookup[error_text] = error_name
    return lookup


_REVERSE_NAME_LOOKUP: Final[Dict[str, StatesErrorNameType]] = (
    _reverse_error_name_lookup()
)



import datetime
from typing import Final

from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalExpression,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment

TIMESTAMP_FORMAT: str = "%Y-%m-%dT%H:%M:%SZ"
# TODO: could be a bit more exact (e.g. 90 shouldn't be a valid minute)
TIMESTAMP_PATTERN: str = r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$"


def parse_timestamp(timestamp: str) -> datetime.datetime:
    # TODO: need to fix this like we're doing for TimestampPath & add a test
    return datetime.datetime.strptime(timestamp, TIMESTAMP_FORMAT)


class Timestamp(WaitFunction):
    # Timestamp
    # An absolute time to state_wait until beginning the state specified in the Next field.
    # Timestamps must conform to the RFC3339 profile of ISO 8601, with the further
    # restrictions that an uppercase T must separate the date and time portions, and
    # an uppercase Z must denote that a numeric time zone offset is not present, for
    # example, 2016-08-18T17:33:00Z.
    # Note
    # Currently, if you specify the state_wait time as a timestamp, Step Functions considers
    # the time value up to seconds and truncates milliseconds.

    def __init__(self, timestamp_literal: str):
        self.timestamp: datetime.datetime = parse_timestamp(timestamp_literal)

    def _get_wait_seconds(self, env: Environment) -> int:
        delta = self.timestamp - datetime.datetime.now()
        delta_sec = int(delta.total_seconds())
        return delta_sec


class TimestampJSONata(WaitFunction):
    jsonata_template_value_terminal_expression: Final[
        JSONataTemplateValueTerminalExpression
    ]

    def __init__(
        self,
        jsonata_template_value_terminal_expression: JSONataTemplateValueTerminalExpression,
    ):
        super().__init__()
        self.jsonata_template_value_terminal_expression = (
            jsonata_template_value_terminal_expression
        )

    def _get_wait_seconds(self, env: Environment) -> int:
        # TODO: add snapshot tests to verify AWS's behaviour about invalid values.
        self.jsonata_template_value_terminal_expression.eval(env=env)
        timestamp_str: str = env.stack.pop()
        timestamp_datetime = parse_timestamp(timestamp_str)
        delta = timestamp_datetime - datetime.datetime.now()
        delta_sec = int(delta.total_seconds())
        return delta_sec



"""Implement models for EFS resources.

See AWS docs for details:
https://docs.aws.amazon.com/efs/latest/ug/whatisefs.html
"""

import json
import time
from copy import deepcopy
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple, Union

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import CloudFormationModel
from moto.core.utils import camelcase_to_underscores, underscores_to_camelcase
from moto.ec2 import ec2_backends
from moto.ec2.exceptions import InvalidSubnetIdError
from moto.ec2.models.elastic_network_interfaces import NetworkInterface
from moto.ec2.models.subnets import Subnet
from moto.efs.exceptions import (
    AccessPointNotFound,
    BadRequest,
    FileSystemAlreadyExists,
    FileSystemInUse,
    FileSystemNotFound,
    MountTargetConflict,
    MountTargetNotFound,
    PolicyNotFound,
    SecurityGroupLimitExceeded,
    SecurityGroupNotFound,
    SubnetNotFound,
)
from moto.moto_api._internal import mock_random
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition, md5_hash


def _lookup_az_id(account_id: str, az_name: str) -> Optional[str]:
    """Find the Availability zone ID given the AZ name."""
    ec2 = ec2_backends[account_id][az_name[:-1]]
    for zone in ec2.describe_availability_zones():
        if zone.name == az_name:
            return zone.zone_id
    return None


class AccessPoint(CloudFormationModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        client_token: str,
        file_system_id: str,
        name: Optional[str],
        posix_user: Dict[str, Any],
        root_directory: Dict[str, str],
        context: "EFSBackend",
    ):
        self.access_point_id = f"fsap-{mock_random.get_random_hex(8)}"
        self.access_point_arn = f"arn:{get_partition(region_name)}:elasticfilesystem:{region_name}:{account_id}:access-point/{self.access_point_id}"
        self.client_token = client_token
        self.file_system_id = file_system_id
        self.name = name
        self.posix_user = posix_user
        self.account_id = account_id

        if not root_directory:
            root_directory = {"Path": "/"}

        self.root_directory = root_directory
        self.context = context

    def info_json(self) -> Dict[str, Any]:
        tags = self.context.list_tags_for_resource(self.access_point_id)
        return {
            "ClientToken": self.client_token,
            "Name": self.name,
            "Tags": tags,
            "AccessPointId": self.access_point_id,
            "AccessPointArn": self.access_point_arn,
            "FileSystemId": self.file_system_id,
            "PosixUser": self.posix_user,
            "RootDirectory": self.root_directory,
            "OwnerId": self.account_id,
            "LifeCycleState": "available",
        }

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::EFS::AccessPoint"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "AccessPoint":
        props = cloudformation_json["Properties"]
        file_system_id = props["FileSystemId"]
        posix_user = props.get("PosixUser", {})
        root_directory = props.get("RootDirectory", {})
        tags = props.get("AccessPointTags", [])

        backend: EFSBackend = efs_backends[account_id][region_name]
        return backend.create_access_point(
            resource_name,
            file_system_id=file_system_id,
            posix_user=posix_user,
            root_directory=root_directory,
            tags=tags,
        )

    def delete(self, account_id: str, region_name: str) -> None:
        backend: EFSBackend = efs_backends[account_id][region_name]
        backend.delete_access_point(self.access_point_id)


class FileSystem(CloudFormationModel):
    """A model for an EFS File System Volume."""

    def __init__(
        self,
        account_id: str,
        region_name: str,
        creation_token: str,
        file_system_id: str,
        context: "EFSBackend",
        performance_mode: str,
        encrypted: bool,
        kms_key_id: str,
        throughput_mode: str,
        provisioned_throughput_in_mibps: int,
        availability_zone_name: str,
        backup: bool,
    ):
        if availability_zone_name:
            backup = True
        if kms_key_id and not encrypted:
            raise BadRequest('If kms_key_id given, "encrypted" must be True.')

        # Save given parameters
        self.creation_token = creation_token
        self.performance_mode = performance_mode or "generalPurpose"
        self.encrypted = encrypted or False
        self.kms_key_id = kms_key_id
        self.throughput_mode = throughput_mode or "bursting"
        self.provisioned_throughput_in_mibps = provisioned_throughput_in_mibps
        self.availability_zone_name = availability_zone_name
        self.availability_zone_id = None
        if self.availability_zone_name:
            self.availability_zone_id = _lookup_az_id(
                account_id, self.availability_zone_name
            )
        self._backup = backup
        self.lifecycle_policies: List[Dict[str, str]] = []
        self.file_system_policy: Optional[str] = None

        self._context = context

        # Generate AWS-assigned parameters
        self.file_system_id = file_system_id
        self.file_system_arn = f"arn:{get_partition(region_name)}:elasticfilesystem:{region_name}:{account_id}:file-system/{self.file_system_id}"
        self.creation_time = time.time()
        self.owner_id = account_id

        # Initialize some state parameters
        self.life_cycle_state = "available"
        self._mount_targets: Dict[str, MountTarget] = {}
        self._size_value = 0

    @property
    def size_in_bytes(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {
            "Value": self._size_value,
            "ValueInIA": 0,
            "ValueInStandard": self._size_value,
            "Timestamp": time.time(),
        }

    @property
    def physical_resource_id(self) -> str:
        return self.file_system_id

    @property
    def number_of_mount_targets(self) -> int:
        return len(self._mount_targets)

    @property
    def backup_policy(self) -> Optional[Dict[str, str]]:
        if self._backup:
            return {"Status": "ENABLED"}
        else:
            return None

    def info_json(self) -> Dict[str, Any]:
        ret = {
            underscores_to_camelcase(k.capitalize()): v
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }
        tags = self._context.list_tags_for_resource(self.file_system_id)
        name = ""
        for tag in tags:
            if tag["Key"] == "Name":
                name = tag["Value"]
                break

        ret.update(
            Tags=tags,
            SizeInBytes=self.size_in_bytes,
            NumberOfMountTargets=self.number_of_mount_targets,
            Name=name,
        )
        return ret

    def add_mount_target(self, subnet: Subnet, mount_target: "MountTarget") -> None:
        # Check that the mount target doesn't violate constraints.
        for other_mount_target in self._mount_targets.values():
            if other_mount_target.subnet_vpc_id != subnet.vpc_id:
                raise MountTargetConflict(
                    "requested subnet for new mount target is not in the same VPC as existing mount targets"
                )

        if subnet.availability_zone in self._mount_targets:
            raise MountTargetConflict("mount target already exists in this AZ")

        self._mount_targets[subnet.availability_zone] = mount_target

    def has_mount_target(self, subnet: Subnet) -> bool:
        return subnet.availability_zone in self._mount_targets

    def iter_mount_targets(self) -> Iterator["MountTarget"]:
        for mt in self._mount_targets.values():
            yield mt

    def remove_mount_target(self, subnet: Subnet) -> None:
        del self._mount_targets[subnet.availability_zone]

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::EFS::FileSystem"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "FileSystem":
        props = cloudformation_json["Properties"]
        performance_mode = props.get("PerformanceMode", "generalPurpose")
        encrypted = props.get("Encrypted", False)
        kms_key_id = props.get("KmsKeyId")
        throughput_mode = props.get("ThroughputMode", "bursting")
        provisioned_throughput_in_mibps = props.get("ProvisionedThroughputInMibps", 0)
        availability_zone_name = props.get("AvailabilityZoneName")
        backup = props.get("BackupPolicy", {}).get("Status") == "ENABLED"
        tags = props.get("FileSystemTags", [])

        lifecycle_policies = props.get("LifecyclePolicies", [])

        backend: EFSBackend = efs_backends[account_id][region_name]
        fs = backend.create_file_system(
            resource_name,
            performance_mode=performance_mode,
            encrypted=encrypted,
            kms_key_id=kms_key_id,
            throughput_mode=throughput_mode,
            provisioned_throughput_in_mibps=provisioned_throughput_in_mibps,
            availability_zone_name=availability_zone_name,
            backup=backup,
            tags=tags,
        )

        if lifecycle_policies:
            backend.put_lifecycle_configuration(
                file_system_id=fs.file_system_id, policies=lifecycle_policies
            )
        return fs

    @classmethod
    def delete_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
    ) -> None:
        efs_backends[account_id][region_name].delete_file_system(resource_name)


class MountTarget(CloudFormationModel):
    """A model for an EFS Mount Target."""

    def __init__(
        self,
        account_id: str,
        file_system: FileSystem,
        subnet: Subnet,
        ip_address: Optional[str],
        security_groups: Optional[List[str]],
    ):
        # Set the simple given parameters.
        self.file_system_id = file_system.file_system_id
        self._file_system = file_system
        self._file_system.add_mount_target(subnet, self)
        self.subnet_id = subnet.id
        self._subnet = subnet
        self.vpc_id = subnet.vpc_id
        self.security_groups = security_groups

        # Check the number of security groups.
        if self.security_groups is not None and len(self.security_groups) > 5:
            raise SecurityGroupLimitExceeded(
                "The maximum number of security groups per interface has been reached."
            )

        # Get an IP address if needed, otherwise validate the one we're given.
        if ip_address is None:
            ip_address = subnet.get_available_subnet_ip(self)  # type: ignore[arg-type]
        else:
            try:
                subnet.request_ip(ip_address, self)  # type: ignore[arg-type]
            except Exception as e:
                if "IP" in str(e) and "CIDR" in str(e):
                    raise BadRequest(
                        "Address does not fall within the subnet's address range"
                    )
                else:
                    raise e
        self.ip_address = ip_address

        # Init non-user-assigned values.
        self.owner_id = account_id
        self.mount_target_id = f"fsmt-{mock_random.get_random_hex()}"
        self.life_cycle_state = "available"
        self.network_interface_id: Optional[str] = None
        self.availability_zone_id = subnet.availability_zone_id
        self.availability_zone_name = subnet.availability_zone

    def clean_up(self) -> None:
        self._file_system.remove_mount_target(self._subnet)
        self._subnet.del_subnet_ip(self.ip_address)

    def set_network_interface(self, network_interface: NetworkInterface) -> None:
        self.network_interface_id = network_interface.id

    def info_json(self) -> Dict[str, Any]:
        return {
            underscores_to_camelcase(k.capitalize()): v
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }

    @property
    def physical_resource_id(self) -> str:
        return self.mount_target_id

    @property
    def subnet_vpc_id(self) -> str:
        return self._subnet.vpc_id

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        return "AWS::EFS::MountTarget"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "MountTarget":
        props = deepcopy(cloudformation_json["Properties"])
        props = {camelcase_to_underscores(k): v for k, v in props.items()}
        return efs_backends[account_id][region_name].create_mount_target(**props)

    @classmethod
    def update_from_cloudformation_json(  # type: ignore[misc]
        cls,
        original_resource: Any,
        new_resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
    ) -> None:
        raise NotImplementedError(
            "Updates of EFS Mount Target via cloudformation are not yet implemented."
        )

    @classmethod
    def delete_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
    ) -> None:
        efs_backends[account_id][region_name].delete_mount_target(resource_name)


class EFSBackend(BaseBackend):
    """The backend manager of EFS resources.

    This is the state-machine for each region, tracking the file systems, mount targets,
    and eventually access points that are deployed. Creating, updating, and destroying
    such resources should always go through this class.
    """

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.creation_tokens: Set[str] = set()
        self.access_points: Dict[str, AccessPoint] = dict()
        self.file_systems_by_id: Dict[str, FileSystem] = {}
        self.mount_targets_by_id: Dict[str, MountTarget] = {}
        self.next_markers: Dict[str, Union[List[MountTarget], List[FileSystem]]] = {}
        self.tagging_service = TaggingService()

    def _mark_description(
        self, corpus: Union[List[MountTarget], List[FileSystem]], max_items: int
    ) -> Optional[str]:
        if max_items < len(corpus):
            new_corpus = corpus[max_items:]
            new_corpus_dict = [c.info_json() for c in new_corpus]
            new_hash = md5_hash(json.dumps(new_corpus_dict).encode("utf-8"))
            next_marker = new_hash.hexdigest()
            self.next_markers[next_marker] = new_corpus
        else:
            next_marker = None
        return next_marker

    @property
    def ec2_backend(self) -> Any:  # type: ignore[misc]
        return ec2_backends[self.account_id][self.region_name]

    def create_file_system(
        self,
        creation_token: str,
        performance_mode: str,
        encrypted: bool,
        kms_key_id: str,
        throughput_mode: str,
        provisioned_throughput_in_mibps: int,
        availability_zone_name: str,
        backup: bool,
        tags: List[Dict[str, str]],
    ) -> FileSystem:
        """Create a new EFS File System Volume.

        https://docs.aws.amazon.com/efs/latest/ug/API_CreateFileSystem.html
        """
        if not creation_token:
            raise ValueError("No creation token given.")
        if creation_token in self.creation_tokens:
            raise FileSystemAlreadyExists(creation_token)

        # Create a new file system ID:
        def make_id() -> str:
            return f"fs-{mock_random.get_random_hex()}"

        fsid = make_id()
        while fsid in self.file_systems_by_id:
            fsid = make_id()
        self.file_systems_by_id[fsid] = FileSystem(
            self.account_id,
            self.region_name,
            creation_token,
            fsid,
            context=self,
            performance_mode=performance_mode,
            encrypted=encrypted,
            kms_key_id=kms_key_id,
            throughput_mode=throughput_mode,
            provisioned_throughput_in_mibps=provisioned_throughput_in_mibps,
            availability_zone_name=availability_zone_name,
            backup=backup,
        )
        self.tag_resource(fsid, tags)
        self.creation_tokens.add(creation_token)
        return self.file_systems_by_id[fsid]

    def describe_file_systems(
        self,
        marker: Optional[str] = None,
        max_items: int = 10,
        creation_token: Optional[str] = None,
        file_system_id: Optional[str] = None,
    ) -> Tuple[Optional[str], List[FileSystem]]:
        """Describe all the EFS File Systems, or specific File Systems.

        https://docs.aws.amazon.com/efs/latest/ug/API_DescribeFileSystems.html
        """
        # Restrict the possible corpus of results based on inputs.
        if creation_token and file_system_id:
            raise BadRequest(
                "Request cannot contain both a file system ID and a creation token."
            )
        elif creation_token:
            # Handle the creation token case.
            corpus = []
            for fs in self.file_systems_by_id.values():
                if fs.creation_token == creation_token:
                    corpus.append(fs)
        elif file_system_id:
            # Handle the case that a file_system_id is given.
            if file_system_id not in self.file_systems_by_id:
                raise FileSystemNotFound(file_system_id)
            corpus = [self.file_systems_by_id[file_system_id]]
        elif marker is not None:
            # Handle the case that a marker is given.
            if marker not in self.next_markers:
                raise BadRequest("Invalid Marker")
            corpus = self.next_markers[marker]  # type: ignore[assignment]
        else:
            # Handle the vanilla case.
            corpus = [fs for fs in self.file_systems_by_id.values()]

        # Handle the max_items parameter.
        file_systems = corpus[:max_items]
        next_marker = self._mark_description(corpus, max_items)
        return next_marker, file_systems

    def create_mount_target(
        self,
        file_system_id: str,
        subnet_id: str,
        ip_address: Optional[str] = None,
        security_groups: Optional[List[str]] = None,
    ) -> MountTarget:
        """Create a new EFS Mount Target for a given File System to a given subnet.

        Note that you can only create one mount target for each availability zone
        (which is implied by the subnet ID).

        https://docs.aws.amazon.com/efs/latest/ug/API_CreateMountTarget.html
        """
        # Get the relevant existing resources
        try:
            subnet = self.ec2_backend.get_subnet(subnet_id)
        except InvalidSubnetIdError:
            raise SubnetNotFound(subnet_id)
        if file_system_id not in self.file_systems_by_id:
            raise FileSystemNotFound(file_system_id)
        file_system = self.file_systems_by_id[file_system_id]

        # Validate the security groups.
        if security_groups:
            sg_lookup = {sg.id for sg in self.ec2_backend.describe_security_groups()}
            for sg_id in security_groups:
                if sg_id not in sg_lookup:
                    raise SecurityGroupNotFound(sg_id)

        # Create the new mount target
        mount_target = MountTarget(
            self.account_id, file_system, subnet, ip_address, security_groups
        )

        # Establish the network interface.
        network_interface = self.ec2_backend.create_network_interface(
            subnet, [mount_target.ip_address], group_ids=security_groups
        )
        mount_target.set_network_interface(network_interface)

        # Record the new mount target
        self.mount_targets_by_id[mount_target.mount_target_id] = mount_target
        return mount_target

    def describe_mount_targets(
        self,
        max_items: int,
        file_system_id: Optional[str],
        mount_target_id: Optional[str],
        access_point_id: Optional[str],
        marker: Optional[str],
    ) -> Tuple[Optional[str], List[MountTarget]]:
        """Describe the mount targets given an access point ID, mount target ID or a file system ID.

        https://docs.aws.amazon.com/efs/latest/ug/API_DescribeMountTargets.html
        """
        # Restrict the possible corpus of results based on inputs.
        if not (bool(file_system_id) ^ bool(mount_target_id) ^ bool(access_point_id)):
            raise BadRequest("Must specify exactly one mutually exclusive parameter.")

        if access_point_id:
            file_system_id = self.access_points[access_point_id].file_system_id

        if file_system_id:
            # Handle the case that a file_system_id is given.
            if file_system_id not in self.file_systems_by_id:
                raise FileSystemNotFound(file_system_id)
            corpus = [
                mt
                for mt in self.file_systems_by_id[file_system_id].iter_mount_targets()
            ]
        elif mount_target_id:
            if mount_target_id not in self.mount_targets_by_id:
                raise MountTargetNotFound(mount_target_id)
            # Handle mount target specification case.
            corpus = [self.mount_targets_by_id[mount_target_id]]

        # Handle the case that a marker is given. Note that the handling is quite
        # different from that in describe_file_systems.
        if marker is not None:
            if marker not in self.next_markers:
                raise BadRequest("Invalid Marker")
            corpus_mtids = {m.mount_target_id for m in corpus}
            marked_mtids = {m.mount_target_id for m in self.next_markers[marker]}  # type: ignore[union-attr]
            mt_ids = corpus_mtids & marked_mtids
            corpus = [self.mount_targets_by_id[mt_id] for mt_id in mt_ids]

        # Handle the max_items parameter.
        mount_targets = corpus[:max_items]
        next_marker = self._mark_description(corpus, max_items)
        return next_marker, mount_targets

    def delete_file_system(self, file_system_id: str) -> None:
        """Delete the file system specified by the given file_system_id.

        Note that mount targets must be deleted first.

        https://docs.aws.amazon.com/efs/latest/ug/API_DeleteFileSystem.html
        """
        if file_system_id not in self.file_systems_by_id:
            raise FileSystemNotFound(file_system_id)

        file_system = self.file_systems_by_id[file_system_id]
        if file_system.number_of_mount_targets > 0:
            raise FileSystemInUse(
                "Must delete all mount targets before deleting file system."
            )

        del self.file_systems_by_id[file_system_id]
        self.creation_tokens.remove(file_system.creation_token)

    def delete_mount_target(self, mount_target_id: str) -> None:
        """Delete a mount target specified by the given mount_target_id.

        Note that this will also delete a network interface.

        https://docs.aws.amazon.com/efs/latest/ug/API_DeleteMountTarget.html
        """
        if mount_target_id not in self.mount_targets_by_id:
            raise MountTargetNotFound(mount_target_id)

        mount_target = self.mount_targets_by_id[mount_target_id]
        self.ec2_backend.delete_network_interface(mount_target.network_interface_id)
        del self.mount_targets_by_id[mount_target_id]
        mount_target.clean_up()

    def describe_backup_policy(self, file_system_id: str) -> Dict[str, str]:
        backup_policy = self.file_systems_by_id[file_system_id].backup_policy
        if not backup_policy:
            raise PolicyNotFound("None")
        return backup_policy

    def put_lifecycle_configuration(
        self, file_system_id: str, policies: List[Dict[str, str]]
    ) -> None:
        _, fss = self.describe_file_systems(file_system_id=file_system_id)
        file_system = fss[0]
        file_system.lifecycle_policies = policies

    def describe_lifecycle_configuration(
        self, file_system_id: str
    ) -> List[Dict[str, str]]:
        _, fss = self.describe_file_systems(file_system_id=file_system_id)
        file_system = fss[0]
        return file_system.lifecycle_policies

    def describe_mount_target_security_groups(
        self, mount_target_id: str
    ) -> Optional[List[str]]:
        if mount_target_id not in self.mount_targets_by_id:
            raise MountTargetNotFound(mount_target_id)

        mount_target = self.mount_targets_by_id[mount_target_id]
        return mount_target.security_groups

    def modify_mount_target_security_groups(
        self, mount_target_id: str, security_groups: List[str]
    ) -> None:
        if mount_target_id not in self.mount_targets_by_id:
            raise MountTargetNotFound(mount_target_id)

        mount_target = self.mount_targets_by_id[mount_target_id]
        mount_target.security_groups = security_groups

        self.ec2_backend.modify_network_interface_attribute(
            eni_id=mount_target.network_interface_id, group_ids=security_groups
        )

    def create_access_point(
        self,
        client_token: str,
        tags: List[Dict[str, str]],
        file_system_id: str,
        posix_user: Dict[str, Any],
        root_directory: Dict[str, Any],
    ) -> AccessPoint:
        name = next((tag["Value"] for tag in tags if tag["Key"] == "Name"), None)
        access_point = AccessPoint(
            self.account_id,
            self.region_name,
            client_token,
            file_system_id,
            name,
            posix_user,
            root_directory,
            context=self,
        )
        self.tagging_service.tag_resource(access_point.access_point_id, tags)
        self.access_points[access_point.access_point_id] = access_point
        return access_point

    def describe_access_points(
        self,
        access_point_id: Optional[str],
        file_system_id: Optional[str],
    ) -> List[AccessPoint]:
        """
        Pagination is not yet implemented
        """
        if access_point_id:
            if access_point_id not in self.access_points:
                raise AccessPointNotFound(access_point_id)
            return [self.access_points[access_point_id]]
        elif file_system_id:
            return [
                access_point
                for access_point in self.access_points.values()
                if access_point.file_system_id == file_system_id
            ]
        return list(self.access_points.values())

    def delete_access_point(self, access_point_id: str) -> None:
        self.access_points.pop(access_point_id, None)

    def list_tags_for_resource(self, resource_id: str) -> List[Dict[str, str]]:
        return self.tagging_service.list_tags_for_resource(resource_id)["Tags"]

    def tag_resource(self, resource_id: str, tags: List[Dict[str, str]]) -> None:
        self.tagging_service.tag_resource(resource_id, tags)

    def untag_resource(self, resource_id: str, tag_keys: List[str]) -> None:
        self.tagging_service.untag_resource_using_names(resource_id, tag_keys)

    def describe_file_system_policy(self, file_system_id: str) -> str:
        policy = self.file_systems_by_id[file_system_id].file_system_policy
        if not policy:
            raise PolicyNotFound("None")
        return policy

    def put_file_system_policy(
        self, file_system_id: str, policy: str, bypass_policy_lockout_safety_check: bool
    ) -> None:
        file_system = self.file_systems_by_id[file_system_id]
        file_system.file_system_policy = policy


efs_backends = BackendDict(EFSBackend, "efs")



import copy
from typing import Final, Optional

from jsonpath_ng import parse

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ResultPath(EvalComponent):
    DEFAULT_PATH: Final[str] = "$"

    result_path_src: Final[Optional[str]]

    def __init__(self, result_path_src: Optional[str]):
        self.result_path_src = result_path_src

    def _eval_body(self, env: Environment) -> None:
        state_input = env.states.get_input()

        # Discard task output if there is one, and set the output ot be the state's input.
        if self.result_path_src is None:
            env.stack.clear()
            env.stack.append(state_input)
            return

        # Transform the output with the input.
        current_output = env.stack.pop()
        result_expr = parse(self.result_path_src)
        state_output = result_expr.update_or_create(
            state_input, copy.deepcopy(current_output)
        )
        env.stack.append(state_output)



"""Handles incoming quicksight requests, invokes methods, returns responses."""

import json
from urllib.parse import unquote

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse

from .models import QuickSightBackend, quicksight_backends


class QuickSightResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="quicksight")

    @property
    def quicksight_backend(self) -> QuickSightBackend:
        """Return backend instance specific for this region."""
        return quicksight_backends[self.current_account][self.region]

    def create_data_set(self) -> str:
        params = json.loads(self.body)
        data_set_id = params.get("DataSetId")
        name = params.get("Name")
        data_set = self.quicksight_backend.create_data_set(data_set_id, name)
        return json.dumps(data_set.to_json())

    def create_group(self) -> str:
        params = json.loads(self.body)
        group_name = params.get("GroupName")
        description = params.get("Description")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group = self.quicksight_backend.create_group(
            group_name=group_name,
            description=description,
            aws_account_id=aws_account_id,
            namespace=namespace,
        )
        return json.dumps(dict(Group=group.to_json()))

    def create_group_membership(self) -> str:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group_name = unquote(self._get_param("GroupName"))
        member_name = unquote(self._get_param("MemberName"))
        member = self.quicksight_backend.create_group_membership(
            aws_account_id, namespace, group_name, member_name
        )
        return json.dumps({"GroupMember": member.to_json()})

    def create_ingestion(self) -> str:
        data_set_id = self._get_param("DataSetId")
        ingestion_id = self._get_param("IngestionId")
        ingestion = self.quicksight_backend.create_ingestion(data_set_id, ingestion_id)
        return json.dumps(ingestion.to_json())

    def describe_group_membership(self) -> str:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group_name = unquote(self._get_param("GroupName"))
        member_name = unquote(self._get_param("MemberName"))
        member = self.quicksight_backend.describe_group_membership(
            aws_account_id, namespace, group_name, member_name
        )
        return json.dumps({"GroupMember": member.to_json()})

    def list_groups(self) -> str:
        max_results = self._get_int_param("max-results")
        next_token = self._get_param("next-token")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        groups, next_token = self.quicksight_backend.list_groups(
            aws_account_id, namespace, max_results=max_results, next_token=next_token
        )
        return json.dumps(
            {"NextToken": next_token, "GroupList": [g.to_json() for g in groups]}
        )

    def list_group_memberships(self) -> str:
        max_results = self._get_int_param("max-results")
        next_token = self._get_param("next-token")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group_name = unquote(self._get_param("GroupName"))
        members, next_token = self.quicksight_backend.list_group_memberships(
            aws_account_id,
            namespace,
            group_name,
            max_results=max_results,
            next_token=next_token,
        )
        return json.dumps(
            {"NextToken": next_token, "GroupMemberList": [m.to_json() for m in members]}
        )

    def list_users(self) -> str:
        max_results = self._get_int_param("max-results")
        next_token = self._get_param("next-token")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        users, next_token = self.quicksight_backend.list_users(
            aws_account_id, namespace, max_results=max_results, next_token=next_token
        )
        return json.dumps(
            {"NextToken": next_token, "UserList": [u.to_json() for u in users]}
        )

    def list_user_groups(self) -> str:
        max_results = self._get_int_param("max-results")
        next_token = self._get_param("next-token")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        user_name = unquote(self._get_param("UserName"))
        groups, next_token = self.quicksight_backend.list_user_groups(
            aws_account_id,
            namespace,
            user_name,
            max_results=max_results,
            next_token=next_token,
        )
        return json.dumps(
            {"NextToken": next_token, "GroupList": [g.to_json() for g in groups]}
        )

    def register_user(self) -> str:
        params = json.loads(self.body)
        identity_type = params.get("IdentityType")
        email = params.get("Email")
        user_role = params.get("UserRole")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        user_name = params.get("UserName")
        user = self.quicksight_backend.register_user(
            identity_type=identity_type,
            email=email,
            user_role=user_role,
            aws_account_id=aws_account_id,
            namespace=namespace,
            user_name=user_name,
        )
        return json.dumps(dict(User=user.to_json(), UserInvitationUrl="TBD"))

    def update_user(self) -> str:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        user_name = unquote(self._get_param("UserName"))
        body = json.loads(self.body)
        email = body.get("Email", None)
        user_role = body.get("Role", None)

        user = self.quicksight_backend.update_user(
            aws_account_id, namespace, user_name, email, user_role
        )
        return json.dumps(dict(User=user.to_json()))

    def describe_group(self) -> str:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group_name = unquote(self._get_param("GroupName"))

        group = self.quicksight_backend.describe_group(
            aws_account_id, namespace, group_name
        )
        return json.dumps(dict(Group=group.to_json()))

    def describe_user(self) -> str:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        user_name = unquote(self._get_param("UserName"))

        user = self.quicksight_backend.describe_user(
            aws_account_id, namespace, user_name
        )
        return json.dumps(dict(User=user.to_json()))

    def delete_group(self) -> TYPE_RESPONSE:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group_name = unquote(self._get_param("GroupName"))

        self.quicksight_backend.delete_group(aws_account_id, namespace, group_name)
        return 204, {"status": 204}, json.dumps({"Status": 204})

    def delete_user(self) -> TYPE_RESPONSE:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        user_name = unquote(self._get_param("UserName"))

        self.quicksight_backend.delete_user(aws_account_id, namespace, user_name)
        return 204, {"status": 204}, json.dumps({"Status": 204})

    def update_group(self) -> str:
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        group_name = unquote(self._get_param("GroupName"))
        description = json.loads(self.body).get("Description")

        group = self.quicksight_backend.update_group(
            aws_account_id, namespace, group_name, description
        )
        return json.dumps(dict(Group=group.to_json()))

    def search_groups(self) -> str:
        max_results = self._get_int_param("max-results")
        next_token = self._get_param("next-token")
        aws_account_id = self._get_param("AwsAccountId")
        namespace = self._get_param("Namespace")
        body = json.loads(self.body)

        groups, next_token = self.quicksight_backend.search_groups(
            aws_account_id,
            namespace,
            body.get("Filters", None),
            max_results=max_results,
            next_token=next_token,
        )
        return json.dumps(
            {"NextToken": next_token, "GroupList": [g.to_json() for g in groups]}
        )



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.moto_api._internal import mock_random

# The log group must be in the us-east-1 region.
TEST_REGION = "us-east-1"


def create_hosted_zone_id(route53_client, hosted_zone_test_name):
    """Return ID of a newly created Route53 public hosted zone"""
    response = route53_client.create_hosted_zone(
        Name=hosted_zone_test_name,
        CallerReference=f"test_caller_ref_{mock_random.get_random_hex(6)}",
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 201
    assert "HostedZone" in response and response["HostedZone"]["Id"]
    return response["HostedZone"]["Id"]


def create_log_group_arn(logs_client, hosted_zone_test_name):
    """Return ARN of a newly created CloudWatch log group."""
    log_group_name = f"/aws/route53/{hosted_zone_test_name}"
    response = logs_client.create_log_group(logGroupName=log_group_name)
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    log_group_arn = None
    response = logs_client.describe_log_groups()
    for entry in response["logGroups"]:
        if entry["logGroupName"] == log_group_name:
            log_group_arn = entry["arn"]
            break
    return log_group_arn


@mock_aws
def test_create_query_logging_config_bad_args():
    """Test bad arguments to create_query_logging_config()."""
    client = boto3.client("route53", region_name=TEST_REGION)
    logs_client = boto3.client("logs", region_name=TEST_REGION)

    hosted_zone_test_name = f"route53_query_log_{mock_random.get_random_hex(6)}.test"
    hosted_zone_id = create_hosted_zone_id(client, hosted_zone_test_name)
    log_group_arn = create_log_group_arn(logs_client, hosted_zone_test_name)

    # Check exception:  NoSuchHostedZone
    with pytest.raises(ClientError) as exc:
        client.create_query_logging_config(
            HostedZoneId="foo", CloudWatchLogsLogGroupArn=log_group_arn
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchHostedZone"
    assert "No hosted zone found with ID: foo" in err["Message"]

    # Check exception:  InvalidInput (bad CloudWatch Logs log ARN)
    with pytest.raises(ClientError) as exc:
        client.create_query_logging_config(
            HostedZoneId=hosted_zone_id,
            CloudWatchLogsLogGroupArn=f"arn:aws:logs:{TEST_REGION}:{ACCOUNT_ID}:foo-bar:foo",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"
    assert "The ARN for the CloudWatch Logs log group is invalid" in err["Message"]

    # Check exception:  InvalidInput (CloudWatch Logs log not in us-east-1)
    with pytest.raises(ClientError) as exc:
        client.create_query_logging_config(
            HostedZoneId=hosted_zone_id,
            CloudWatchLogsLogGroupArn=log_group_arn.replace(TEST_REGION, "us-west-1"),
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"
    assert "The ARN for the CloudWatch Logs log group is invalid" in err["Message"]

    # Check exception:  NoSuchCloudWatchLogsLogGroup
    with pytest.raises(ClientError) as exc:
        client.create_query_logging_config(
            HostedZoneId=hosted_zone_id,
            CloudWatchLogsLogGroupArn=log_group_arn.replace(
                hosted_zone_test_name, "foo"
            ),
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchCloudWatchLogsLogGroup"
    assert "The specified CloudWatch Logs log group doesn't exist" in err["Message"]

    # Check exception:  QueryLoggingConfigAlreadyExists
    client.create_query_logging_config(
        HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
    )
    with pytest.raises(ClientError) as exc:
        client.create_query_logging_config(
            HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "QueryLoggingConfigAlreadyExists"
    assert (
        "A query logging configuration already exists for this hosted zone"
        in err["Message"]
    )


@mock_aws
def test_create_query_logging_config_good_args():
    """Test a valid create_logging_config() request."""
    client = boto3.client("route53", region_name=TEST_REGION)
    logs_client = boto3.client("logs", region_name=TEST_REGION)

    hosted_zone_test_name = f"route53_query_log_{mock_random.get_random_hex(6)}.test"
    hosted_zone_id = create_hosted_zone_id(client, hosted_zone_test_name)
    log_group_arn = create_log_group_arn(logs_client, hosted_zone_test_name)

    response = client.create_query_logging_config(
        HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
    )
    config = response["QueryLoggingConfig"]
    assert config["HostedZoneId"] == hosted_zone_id.split("/")[-1]
    assert config["CloudWatchLogsLogGroupArn"] == log_group_arn
    assert config["Id"]

    location = response["Location"]
    assert (
        location
        == f"https://route53.amazonaws.com/2013-04-01/queryloggingconfig/{config['Id']}"
    )


@mock_aws
def test_delete_query_logging_config():
    """Test valid and invalid delete_query_logging_config requests."""
    client = boto3.client("route53", region_name=TEST_REGION)
    logs_client = boto3.client("logs", region_name=TEST_REGION)

    # Create a query logging config that can then be deleted.
    hosted_zone_test_name = f"route53_query_log_{mock_random.get_random_hex(6)}.test"
    hosted_zone_id = create_hosted_zone_id(client, hosted_zone_test_name)
    log_group_arn = create_log_group_arn(logs_client, hosted_zone_test_name)

    query_response = client.create_query_logging_config(
        HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
    )

    # Test the deletion.
    query_id = query_response["QueryLoggingConfig"]["Id"]
    response = client.delete_query_logging_config(Id=query_id)
    # There is no response other than the usual ResponseMetadata.
    assert list(response.keys()) == ["ResponseMetadata"]

    # Test the deletion of a non-existent query logging config, i.e., the
    # one that was just deleted.
    with pytest.raises(ClientError) as exc:
        client.delete_query_logging_config(Id=query_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchQueryLoggingConfig"
    assert "The query logging configuration does not exist" in err["Message"]


@mock_aws
def test_get_query_logging_config():
    """Test valid and invalid get_query_logging_config requests."""
    client = boto3.client("route53", region_name=TEST_REGION)
    logs_client = boto3.client("logs", region_name=TEST_REGION)

    # Create a query logging config that can then be retrieved.
    hosted_zone_test_name = f"route53_query_log_{mock_random.get_random_hex(6)}.test"
    hosted_zone_id = create_hosted_zone_id(client, hosted_zone_test_name)
    log_group_arn = create_log_group_arn(logs_client, hosted_zone_test_name)

    query_response = client.create_query_logging_config(
        HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
    )

    # Test the retrieval.
    query_id = query_response["QueryLoggingConfig"]["Id"]
    response = client.get_query_logging_config(Id=query_id)
    config = response["QueryLoggingConfig"]
    assert config["HostedZoneId"] == hosted_zone_id.split("/")[-1]
    assert config["CloudWatchLogsLogGroupArn"] == log_group_arn
    assert config["Id"]

    # Test the retrieval of a non-existent query logging config.
    with pytest.raises(ClientError) as exc:
        client.get_query_logging_config(Id="1234567890")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchQueryLoggingConfig"
    assert "The query logging configuration does not exist" in err["Message"]


@mock_aws
def test_list_query_logging_configs_bad_args():
    """Test bad arguments to list_query_logging_configs()."""
    client = boto3.client("route53", region_name=TEST_REGION)
    logs_client = boto3.client("logs", region_name=TEST_REGION)

    # Check exception:  NoSuchHostedZone
    with pytest.raises(ClientError) as exc:
        client.list_query_logging_configs(HostedZoneId="foo", MaxResults="10")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchHostedZone"
    assert "No hosted zone found with ID: foo" in err["Message"]

    # Create a couple of query logging configs to work with.
    for _ in range(3):
        hosted_zone_test_name = (
            f"route53_query_log_{mock_random.get_random_hex(6)}.test"
        )
        hosted_zone_id = create_hosted_zone_id(client, hosted_zone_test_name)
        log_group_arn = create_log_group_arn(logs_client, hosted_zone_test_name)
        client.create_query_logging_config(
            HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
        )

    # Retrieve a query logging config, then request more with an invalid token.
    client.list_query_logging_configs(MaxResults="1")
    with pytest.raises(ClientError) as exc:
        client.list_query_logging_configs(NextToken="foo")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidPaginationToken"
    assert (
        "Route 53 can't get the next page of query logging configurations "
        "because the specified value for NextToken is invalid." in err["Message"]
    )


@mock_aws
def test_list_query_logging_configs_good_args():
    """Test valid arguments to list_query_logging_configs()."""
    client = boto3.client("route53", region_name=TEST_REGION)
    logs_client = boto3.client("logs", region_name=TEST_REGION)

    # Test when there are no query logging configs.
    response = client.list_query_logging_configs()
    query_logging_configs = response["QueryLoggingConfigs"]
    assert len(query_logging_configs) == 0

    # Create a couple of query logging configs to work with.
    zone_ids = []
    for _ in range(10):
        hosted_zone_test_name = (
            f"route53_query_log_{mock_random.get_random_hex(6)}.test"
        )
        hosted_zone_id = create_hosted_zone_id(client, hosted_zone_test_name)
        zone_ids.append(hosted_zone_id)

        log_group_arn = create_log_group_arn(logs_client, hosted_zone_test_name)
        client.create_query_logging_config(
            HostedZoneId=hosted_zone_id, CloudWatchLogsLogGroupArn=log_group_arn
        )

    # Verify all 10 of the query logging configs can be retrieved in one go.
    response = client.list_query_logging_configs()
    query_logging_configs = response["QueryLoggingConfigs"]
    assert len(query_logging_configs) == 10
    for idx, query_logging_config in enumerate(query_logging_configs):
        assert query_logging_config["HostedZoneId"] == zone_ids[idx].split("/")[-1]

    # Request only two of the query logging configs and verify there's a
    # next_token.
    response = client.list_query_logging_configs(MaxResults="2")
    assert len(response["QueryLoggingConfigs"]) == 2
    assert response["NextToken"]

    # Request the remaining 8 query logging configs and verify there is
    # no next token.
    response = client.list_query_logging_configs(
        MaxResults="8", NextToken=response["NextToken"]
    )
    assert len(response["QueryLoggingConfigs"]) == 8
    assert "NextToken" not in response



from typing import List

from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_binding import (
    JSONataTemplateBinding,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value import (
    JSONataTemplateValue,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class JSONataTemplateValueObject(JSONataTemplateValue):
    bindings: List[JSONataTemplateBinding]

    def __init__(self, bindings: List[JSONataTemplateBinding]):
        self.bindings = bindings

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(dict())
        for binding in self.bindings:
            binding.eval(env)



"""cloudfront base URL and path."""

from .responses import CloudFrontResponse

url_bases = [
    r"https?://cloudfront\.amazonaws\.com",
    r"https?://cloudfront\.(.+)\.amazonaws\.com",
]
url_paths = {
    "{0}/2020-05-31/distribution$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/distribution/(?P<distribution_id>[^/]+)$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/distribution/(?P<distribution_id>[^/]+)/config$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/distribution/(?P<distribution_id>[^/]+)/invalidation": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/distribution/(?P<distribution_id>[^/]+)/invalidation/(?P<invalidation_id>[^/]+)": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/key-group$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/key-group/(?P<key_name>[^/]+)$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/tagging$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/origin-access-control$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/origin-access-control/(?P<oac_id>[^/]+)$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/origin-access-control/(?P<oac_id>[^/]+)/config$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/public-key$": CloudFrontResponse.dispatch,
    "{0}/2020-05-31/public-key/(?P<key_name>[^/]+)$": CloudFrontResponse.dispatch,
}



import boto3

from moto import mock_aws


@mock_aws
def test_list_tags_for_db_instance():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
        tags={"key1": "value1", "key2": "value2"},
    )

    arn = response["arn"]
    response = client.list_tags_for_resource(resourceArn=arn)
    assert response["tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_tag_resources():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
        tags={"key1": "value1", "key2": "value2"},
    )

    arn = response["arn"]

    # add a tag
    additional_tags = {"key3": "value3"}
    response = client.tag_resource(resourceArn=arn, tags=additional_tags)

    response = client.list_tags_for_resource(resourceArn=arn)
    assert response["tags"] == {"key1": "value1", "key2": "value2", "key3": "value3"}


@mock_aws
def test_untag_resources():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
        tags={"key1": "value1", "key2": "value2"},
    )

    arn = response["arn"]

    # remove tags
    response = client.untag_resource(resourceArn=arn, tagKeys=["key1"])

    response = client.list_tags_for_resource(resourceArn=arn)
    assert response["tags"] == {"key2": "value2"}



from datetime import datetime, timedelta
from typing import List, Tuple

from moto.core.utils import utcnow


def pytest_parametrize_test_create_get_schedule__with_start_date() -> (
    List[Tuple[datetime, datetime]]
):
    now = utcnow()
    timedelta_kwargs = {"days": 1, "hours": 1, "weeks": 1}
    return_ = []
    for k, v in timedelta_kwargs.items():
        to_append = now + timedelta(**{k: v})
        return_.append((to_append, to_append.replace(microsecond=0)))
    else:
        to_append = now.replace(year=now.year + 1)
        return_.append((to_append, to_append.replace(microsecond=0)))
    return return_


def pytest_parametrize_test_create_schedule__exception_with_start_date() -> (
    List[datetime]
):
    now = utcnow()
    timedelta_kwargs = {"days": 1, "minutes": 6, "hours": 1, "weeks": 1}
    return_ = []
    for k, v in timedelta_kwargs.items():
        return_.append(now - timedelta(**{k: v}))
    else:
        return_.append(now.replace(year=now.year - 1))
    return return_



import abc
from typing import Optional

from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class OutputPath(EvalComponent, abc.ABC): ...


class OutputPathBase(OutputPath):
    DEFAULT_PATH: str = "$"

    output_path: Optional[str]

    def __init__(self, output_path: Optional[str]):
        self.output_path = output_path

    def _eval_body(self, env: Environment) -> None:
        if self.output_path is None:
            env.states.reset(input_value=dict())
        else:
            current_output = env.stack.pop()
            state_output = extract_json(self.output_path, current_output)
            env.states.reset(input_value=state_output)


class OutputPathContextObject(OutputPathBase):
    def __init__(self, output_path: str):
        output_path_tail = output_path[1:]
        super().__init__(output_path=output_path_tail)

    def _eval_body(self, env: Environment) -> None:
        env.stack.pop()  # Discards the state output in favour of the context object path.
        value = extract_json(
            self.output_path, env.states.context_object.context_object_data
        )
        env.states.reset(input_value=value)


class OutputPathVar(OutputPath):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        self.variable_sample = variable_sample

    def _eval_body(self, env: Environment) -> None:
        self.variable_sample.eval(env=env)
        value = env.stack.pop()
        env.states.reset(input_value=value)



import abc
import datetime
import threading
from collections import OrderedDict
from typing import Dict, Final, List, Optional, Tuple

from moto.core.utils import iso_8601_datetime_with_milliseconds
from moto.stepfunctions.parser.api import (
    Arn,
    DescribeMapRunOutput,
    LongArn,
    MapRunExecutionCounts,
    MapRunItemCounts,
    MapRunListItem,
    MapRunStatus,
    Timestamp,
)
from moto.stepfunctions.parser.utils import long_uid


class Counter:
    _mutex: Final[threading.Lock]
    _count: int

    def __init__(self):
        self._mutex = threading.Lock()
        self._count = 0

    def offset(self, offset: int) -> None:
        with self._mutex:
            self._count = self._count + offset

    def count(self, increment: int = 1) -> None:
        with self._mutex:
            self._count += increment

    def get(self) -> int:
        return self._count


class ProgressCounter(abc.ABC):
    aborted: Final[Counter]
    failed: Final[Counter]
    pending: Final[Counter]
    results_written: Final[Counter]
    running: Final[Counter]
    succeeded: Final[Counter]
    timed_out: Final[Counter]
    total: Final[Counter]

    def __init__(self):
        self.aborted = Counter()
        self.failed = Counter()
        self.pending = Counter()
        self.results_written = Counter()
        self.running = Counter()
        self.succeeded = Counter()
        self.timed_out = Counter()
        self.total = Counter()


class ExecutionCounter(ProgressCounter):
    def describe(self) -> MapRunExecutionCounts:
        return MapRunExecutionCounts(
            aborted=self.aborted.get(),
            failed=self.failed.get(),
            pending=self.pending.get(),
            resultsWritten=self.results_written.get(),
            running=self.running.get(),
            succeeded=self.succeeded.get(),
            timedOut=self.timed_out.get(),
            total=self.total.get(),
        )


class ItemCounter(ProgressCounter):
    def describe(self) -> MapRunItemCounts:
        return MapRunItemCounts(
            aborted=self.aborted.get(),
            failed=self.failed.get(),
            pending=self.pending.get(),
            resultsWritten=self.results_written.get(),
            running=self.running.get(),
            succeeded=self.succeeded.get(),
            timedOut=self.timed_out.get(),
            total=self.total.get(),
        )


class MapRunRecord:
    update_event: threading.Event
    map_state_machine_arn: Final[
        LongArn
    ]  # This is the original state machine arn plut the map run arn postfix.
    execution_arn: Arn
    map_run_arn: LongArn
    max_concurrency: int
    execution_counter: Final[ExecutionCounter]
    item_counter: Final[ItemCounter]
    start_date: Timestamp
    status: MapRunStatus
    stop_date: Optional[Timestamp]
    # TODO: add support for failure toleration fields.
    tolerated_failure_count: int
    tolerated_failure_percentage: float

    def __init__(
        self,
        state_machine_arn: Arn,
        execution_arn: Arn,
        max_concurrency: int,
        tolerated_failure_count: int,
        tolerated_failure_percentage: float,
        label: Optional[str],
    ):
        self.update_event = threading.Event()
        (
            map_state_machine_arn,
            map_run_arn,
        ) = self._generate_map_run_arns(
            state_machine_arn=state_machine_arn, label=label
        )
        self.map_run_arn = map_run_arn
        self.map_state_machine_arn = map_state_machine_arn
        self.execution_arn = execution_arn
        self.max_concurrency = max_concurrency
        self.execution_counter = ExecutionCounter()
        self.item_counter = ItemCounter()
        self.start_date = datetime.datetime.now(tz=datetime.timezone.utc)
        self.status = MapRunStatus.RUNNING
        self.stop_date = None
        self.tolerated_failure_count = tolerated_failure_count
        self.tolerated_failure_percentage = tolerated_failure_percentage

    @staticmethod
    def _generate_map_run_arns(
        state_machine_arn: Arn, label: Optional[str]
    ) -> Tuple[LongArn, LongArn]:
        # Generate a new MapRunArn given the StateMachineArn, such that:
        # inp: arn:aws:states:<region>:111111111111:stateMachine:<ArnPart_0idx>
        # MRA: arn:aws:states:<region>:111111111111:mapRun:<ArnPart_0idx>/<MapRunArnPart0_0idx>:<MapRunArnPart1_0idx>
        # SMA: arn:aws:states:<region>:111111111111:mapRun:<ArnPart_0idx>/<MapRunArnPart0_0idx>
        map_run_arn = state_machine_arn.replace(":stateMachine:", ":mapRun:")
        part_1 = long_uid() if label is None else label
        map_run_arn = f"{map_run_arn}/{part_1}:{long_uid()}"
        return f"{state_machine_arn}/{part_1}", map_run_arn

    def set_stop(self, status: MapRunStatus):
        self.status = status
        self.stop_date = datetime.datetime.now(tz=datetime.timezone.utc)

    def describe(self) -> DescribeMapRunOutput:
        describe_output = DescribeMapRunOutput(
            mapRunArn=self.map_run_arn,
            executionArn=self.execution_arn,
            status=self.status,
            startDate=iso_8601_datetime_with_milliseconds(self.start_date),
            maxConcurrency=self.max_concurrency,
            toleratedFailurePercentage=self.tolerated_failure_percentage,
            toleratedFailureCount=self.tolerated_failure_count,
            itemCounts=self.item_counter.describe(),
            executionCounts=self.execution_counter.describe(),
        )
        stop_date = self.stop_date
        if stop_date is not None:
            describe_output["stopDate"] = iso_8601_datetime_with_milliseconds(
                self.stop_date
            )
        return describe_output

    def list_item(self) -> MapRunListItem:
        list_item = MapRunListItem(
            executionArn=self.execution_arn,
            mapRunArn=self.map_run_arn,
            stateMachineArn=self.map_state_machine_arn,
            startDate=iso_8601_datetime_with_milliseconds(self.start_date),
        )
        if self.stop_date:
            list_item["stopDate"] = iso_8601_datetime_with_milliseconds(self.stop_date)
        return list_item

    def update(
        self,
        max_concurrency: Optional[int],
        tolerated_failure_count: Optional[int],
        tolerated_failure_percentage: Optional[float],
    ) -> None:
        if max_concurrency is not None:
            self.max_concurrency = max_concurrency
        if tolerated_failure_count is not None:
            self.tolerated_failure_count = tolerated_failure_count
        if tolerated_failure_percentage is not None:
            self.tolerated_failure_percentage = tolerated_failure_percentage
        self.update_event.set()


class MapRunRecordPoolManager:
    _pool: Dict[str, MapRunRecord]

    def __init__(self):
        self._pool = OrderedDict()

    def add(self, map_run_record: MapRunRecord) -> None:
        self._pool[map_run_record.map_run_arn] = map_run_record

    def get(self, map_run_arn: str) -> Optional[MapRunRecord]:
        return self._pool.get(map_run_arn)

    def get_all(self) -> List[MapRunRecord]:
        return list(self._pool.values())
